//
// Module: BadPixPSFCorrector
// Author: Hassan Siddiqui
//
// Description: 
//
// Point source implementation of an Encircled Energy Corrector class
//
// Modifications: See ChangeLog
// 
// $Id: BadPixPSFCorrector.cc,v 1.2 2005/06/16 12:37:26 rsaxton Exp $

/* General */
#include <cmath>

/* SAS-specific */
#include "Tracer.h"
#include "DataFilter.h"

/* Local */
#include "NumericalRange.h"
#include "BadPixPSFCorrector.h"
#include "MissingPixCorrector.h"
#include "TablePixelIterator.h"
#include "CompactTablePixelIterator.h"
#include "DetmapFilter.h"
#include "ArfDataServer.h"

/* 
   Method: BadPixPSFCorrector(string table, unsigned eefactor, 
               PsfModel * psf, unsigned int ccdid)
         : Constructor
*/
BadPixPSFCorrector::BadPixPSFCorrector(const string & table, 
  unsigned eefactor, PsfModel * psf, unsigned int ccdid):
  MissingPixCorrector(eefactor),
  _psf(psf),
  _table(table),
  _ccdid(ccdid)
{
  PROC;

  /* Create a special table iterator for the bad pixel table */
    _ctpxitr = new CompactTablePixelIterator(_table, "RAWX",
                                "RAWY","TYPE","YEXTENT");

  ENDPROC;
}
/* Destructor */
BadPixPSFCorrector::~BadPixPSFCorrector()
{
  PROC;
   delete _detmapfilter;     _detmapfilter      =0;
   delete _pxitr;            _pxitr            =0;
   delete _ctpxitr;          _ctpxitr            =0;
  ENDPROC;
}

/* 
   Method: filterMap()
         : Removes bad pixels outside source region 
*/
void BadPixPSFCorrector::filterMap() 
{
  PROC;

  /* Set the CCD number in the CAL before filtering */
    calServer.state()->ccd()->set(_ccdid);

  /* Filter the bad pixel table so that only elements within the source 
     area remain.  Use the extension name as the local temporary filename
     NB: Cant be done in constructor because needs arfDataServer to 
     be fully set-up first */

    string localbadfile;
    if (_table.substr(0,_table.find("l")+1) == "temp_badcol") {
      localbadfile = strstr(_table.c_str(), "BADPIX");
      localbadfile = "OFFSET"+localbadfile.substr(6,2);
    } else {
      localbadfile = strstr(_table.c_str(), "BADPIX");
    }
    errstr << "BadPixPSFCorrector: filtering map "<<localbadfile
	   << Sas::message(Msg::App, Msg::Verbose);
      
    DetmapFilter * _detmapfilter = new DetmapFilter();
    _pxitr = _detmapfilter->filteredMap(_ctpxitr, localbadfile+".ds:EVENTS", 
                     arfDataServer->specAttributes(), true);

  ENDPROC;
}

double BadPixPSFCorrector::eeFraction( double energy  ) const
{
  PROC;

  /* Set the CCD number in the CAL before filtering */
  calServer.state()->ccd()->set(_ccdid);

  double fracenclosed   = 0;
  double contrib = 0;

      errstr << "BadPixPSFCorrector: Npix: "<<_pxitr->size()
	     << Sas::message(Msg::App, Msg::Verbose);
      
 /* If there are any pixels - loop over them */
  if (_pxitr->size() > 0) {

   /* Calculate the PSF function at this energy */
    _psf->energy(energy);

 /* Loop over each bad pixel */
    FOREACH(_pxitr)
    {
      PixelElement pixel = _pxitr->currentItem();
      
     /* Add contribution to PSF */
      contrib = _psf->int_psf( pixel.xCentre(), 
                               pixel.yCentre(),
                               pixel.xWidth(),
                               pixel.yWidth() );
      
      errstr << "BadPixPSFCorrector: Pixel details  " << pixel << endl
	     << " PSF contribution " << contrib << endl
	     << Sas::message(Msg::App, Msg::Noisy);
      
      fracenclosed += contrib;
    }
  }
 
  errstr << "BadPixPSFCorrector: Fraction of psf enclosed " 
	 << fracenclosed << endl
	 << Sas::message(Msg::App, Msg::Verbose); 

  RETURN fracenclosed;
}
  
int BadPixPSFCorrector::totalPixels() const
{
  PROC;

   /* Calculate the total number of detector pixels covered by the bad 
      pixels */

     //int total = UNT(_pxitr->size() * _pxitr->wcsconvertor()->xWidth() *
      //                            _pxitr->wcsconvertor()->yWidth());

     vector<double> widths = arfDataServer->positionUtils()->detxyWidth(); 
     int total = UNT(_pxitr->size() * widths[0] * widths[1]);
           errstr<<"Widths: "<<widths[0]<<" "<<widths[1]
                   <<"tot for this CCD: "<<total
	     << Sas::message(Msg::App, Msg::Verbose);

  RETURN total;
}
