//
// Module: CalPsfModel
// Author: Hassan Siddiqui
//
// Description: 
//
// Wrapper class around CAL psf model.
//
// Modifications: See ChangeLog
// 
// $Id: CalPsfModel.cc,v 1.4 2006/06/07 14:31:35 rsaxton Exp $

/* SAS */
#include "errstr.h"
#include "Tracer.h"
#include "CalServer.h"
#include "Region.h"

// Local classes
#include "ArfDataServer.h"
#include "CalPsfModel.h"
#include "PositionUtils.h"

/* Constructor */
CalPsfModel::CalPsfModel( double energy,
			  double detxcentre,
			  double detycentre,
                          AccuracyLevel mode,
			  const PositionUtils * const putils):
  _putils(putils),
  _mode(mode),
  _offset(0,0,0),
  _t(0,0),
  _boxShape(ShapeBasics::CIRCLE),
  _fastmode_factor(1.0)
{
  PROC;

  errstr 
    << "Constructing CalPsfModel with energy=" << energy
    << " detxcentre=" << detxcentre
    << " detycentre=" << detycentre
    << " mode=" << _mode
    << Sas::message(Msg::App, Msg::Verbose);


  _offset = _putils->detxy005ToCamCoord2( detxcentre, 
					  detycentre );

  _t = _putils->camCoord2ToTelCoord(_offset);
  
  calServer.state()->accuracyLevel()->set(_mode);

  PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);

  /* Use a nasty trick to convert a reference to a pointer. */
  errstr 
    << "Calling PsfDataServer for a Psf at TELCOORD position ["
    << setprecision(10) << _t.theta()
    << ","
    << setprecision(10) << _t.phi()
    << "]"
    << Sas::message(Msg::App, Msg::Verbose);
  
    Psf & p = psfs->psf(energy, _t.theta(), _t.phi());
  
    _psf = &p;

    /* Reduce bin size */
    if (_psf->binSizeChangeable())
      {

	const unsigned OVERSAMPLE_FACTOR = 4;

      	Psf::BinSize bz = _psf->binSize();
	Psf::BinSize bz_out(bz.binX()/OVERSAMPLE_FACTOR, bz.binY()/OVERSAMPLE_FACTOR);

	errstr 
	  << " Changing psf binsize from ["
	  << bz.binX() << "," << bz.binY()
	  << "] to [" 
	  << bz_out.binX() << "," << bz_out.binY() 
	  << "]"
	  << Sas::message(Msg::App, Msg::Noisy);
	_psf->binSize(bz_out);
      }

 
  ENDPROC;
};

CalPsfModel::~CalPsfModel()
{
  PROC;
  errstr 
    << "CalPsfModel destroyed."
    << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}

/* Return PSF function at given position */
double CalPsfModel::psf( double detx, 
			 double dety) const
{
  PROC;
  CamCoords c = _putils->detxy005ToCamCoord2( detx, 
					      dety );

  double out = (*_psf)(c.x(), c.y());

  errstr 
    << "CalPsfModel::psf: for detx=" << detx
    << " dety=" << dety
    << " get a psf of " << out
    << Sas::message(Msg::App, Msg::Noisy);

  RETURN out;
}

/* Calculate integrated PSF over a rectangular area. */
double CalPsfModel::int_psf( double detxboxcentre, 
			     double detyboxcentre, 
			     double detxboxsize, 
			     double detyboxsize ) const
{
  PROC;

  CamCoords c1 = _putils->detxy005ToCamCoord2( detxboxcentre, 
					       detyboxcentre );

  errstr 
    << "CalPsfModel::int_psf: setting box origin to ["
    << setprecision(10) << c1.x()-_offset.x()
    << ","
     << setprecision(10) << c1.y()-_offset.y()
    << "]"
    << Sas::message(Msg::App, Msg::Noisy);

  _psf->atOrigin(c1.x()-_offset.x(),c1.y()-_offset.y());

  /* Know that camcoords2/det have same origin (0,0).
     So can use same method for scales as well as positions */
  CamCoords c2 = _putils->detxy005ToCamCoord2( detxboxsize, 
					       detyboxsize );

  errstr 
    << "CalPsfModel::int_psf: creating a PsfWindow of "
    << setprecision(10) << c2.x()/2 << ","  
    << setprecision(10) << c2.y()/2
    << Sas::message(Msg::App, Msg::Noisy);

  Psf::PsfWindow pw(c2.x()/2., c2.y()/2);

  double out=0.0;

  
     errstr << "Mode"<<calServer.state()->accuracyLevel()->get()
     << Sas::message(Msg::App, Msg::Noisy);

  if (calServer.state()->accuracyLevel()->get() != EXTENDED) {

     out = _psf->enboxedEnergy(pw);

     errstr 
     << "CalPsfModel::int_psf: for "
     << " energy=" << _psf->E()
     << " theta=" << _t.theta()
     << " phi=" << _t.phi()
     << endl
     << " detxboxcentre=" << detxboxcentre
     << " detyboxcentre=" << detyboxcentre
     << " detxboxsize=" << detxboxsize
     << " detyboxsize=" << detyboxsize
     << endl
     << " derive a PsfWindow of "
     << " halfX=" << pw.halfX() 
     << " halfY=" << pw.halfY()
     << endl
     << " and get an enclosed energy " << out
     << Sas::message(Msg::App, Msg::Noisy);
  
   } else {   // circular or annular source box calculation

/* The above doesn't work very well for circular or annular source
   boxes - so use the following technique 
      Algorithm: Find the total PSF within the outer radius of
      the box, subtract the total PSF from within the inner radius
      and normalise for the box size over the annulus area
*/

   /* Calculate the mean radius of this box */
      double boxRadius = sqrt( (c1.x()-_offset.x()) * (c1.x()-_offset.x())
                        + (c1.y()-_offset.y()) * (c1.y()-_offset.y()) );

   /* This technique doesn't work for pixels right on the source
      centre so make sure radius is bigger than half a box */

      CamCoords bit = _putils->detxy005ToCamCoord2(1,1);

      if (boxRadius < c2.x()/2.0 + bit.x()) {
           boxRadius=c2.x()/2.0 + bit.x();
      }

   /* Find the PSF contribution in an annulus about this radius (in arcsec)  */
      double PsfCont = _psf->encircledEnergy(boxRadius + c2.x()/2.0) -
                       _psf->encircledEnergy(boxRadius - c2.x()/2.0); 

   /* Normalise by area to get the PSF fraction in the box */
      out = PsfCont * c2.x() * c2.y() / ( M_PI *
        ((boxRadius + c2.y()/2.0)*(boxRadius + c2.y() /2.0) -
         (boxRadius - c2.y()/2.0)*(boxRadius - c2.y()/2.0)) );

      errstr<<"Int_psf: " 
            <<"Centre: "<<detxboxcentre<<" "<<detyboxcentre
            <<" Widths: "<<detxboxsize<<" "<<detyboxsize
            <<"Radius: "<<boxRadius
            <<"psccont: "<<PsfCont
            <<"Normalised: "<<out
            << Sas::message(Msg::App, Msg::Noisy);
   }

  RETURN out;
}

/* Set energy */
void CalPsfModel::energy(double energy )
{
  PROC;

  _psf->setParameters(energy, _t.theta(), _t.phi());

  ENDPROC;
}

string CalPsfModel::mode() const
{
  PROC;

  RETURN "Dummy";
}

void CalPsfModel::setRadius( vector<double> radii )
{
  
  PROC;
    for (unsigned i=0; i<radii.size(); i++) {

      errstr
        << "Setting radius "<<i<<" = " <<radii[i]
        << Sas::message(Msg::App, Msg::Verbose);

       _radii.push_back(radii[i]);
    }
      errstr
        << "Finished for i="<<radii.size()
        << Sas::message(Msg::App, Msg::Verbose);
  ENDPROC;
}

void CalPsfModel::setNgridEls(unsigned int /* NGridEls */)
{
  PROC;
  /* Do nothing */
  ENDPROC;
}


double CalPsfModel::rad_psf() const
{
  PROC;

      double ee=0;

      errstr
        << "Getting Encircled energy (energy): "<<_psf->E()
        << " radius "<<_radii[0]
        << Sas::message(Msg::App, Msg::Verbose);

  /* Calculation method depends on source box shape */
      if ( _boxShape == ShapeBasics::CIRCLE ||
           _boxShape == ShapeBasics::ANNULUS) {

         ee = _psf->encircledEnergy(_radii[0]); 

  /* Subtract the encircled energy from within the inner radius if
     this is an annulus */

         if (_radii.size() == 2) {
            ee = _psf->encircledEnergy(_radii[1]) - ee;
         }

      } else {   // BOX

         double ysmall=0.0, ybig=0.0;
         double minor=0.0, major=0.0;

  /* For PN Timing/Burst mode need to treat the length in two pieces
     depending on the exact position of the source */
         Xmm::Mode windowMode = calServer.state()->mode()->get();
         if (windowMode == Xmm::FAST_TIMING ||
                     windowMode == Xmm::FAST_BURST) 
         {
            ModeParameters *modeParam = 0; 
                              modeParam = calServer.getAtom(modeParam);

          /* Not easy to get the exact RAWY position of the source. Nominally
             it is placed at RAWY=190, which is available from the keyword
             SRCPOS in the spectrum top level header - here we hard code it */
            int PNFASTMODE_RAWYPOS = 190;

          /* Calc the distance from nominal source centre to chip border
             in CamCoords (mm) */
            ysmall = (modeParam->windowY0()+modeParam->windowDY()-1
                      - PNFASTMODE_RAWYPOS); // In RAW pixels
            ysmall = _putils->detxy005ToCamCoord2(INT(ysmall*4.1*20.0),1).x(); 

          // Calc the size of what is left of the Y box
            ybig = _radii[1] * 2 - ysmall;

            errstr<<ysmall<<" "<<ybig
                  << Sas::message(Msg::App, Msg::Noisy);

         } else {
            //ybit->push_back(_radii[1]);
            ysmall=_radii[1];
            ybig=_radii[1];
         }

         double xbit= _radii[0];

  /* Work over two sections of the box */
         for (int s=0; s<=1; s++) {

            double y = ysmall;
            if (s==1) {y=ybig;}

    /* Set the minor and major axes */
            if (xbit > y ) {
               major=xbit;
               minor=y;
            } else {
               major=y;
               minor=xbit;
            }

    /* Calculate the largest circle which is completely enclosed */
           double ee_inner = _psf->encircledEnergy(minor);

    /* Find the encircled energy for the circle touched by the corners of
       the rectangle */

           double diagonal = sqrt(minor*minor + major*major);

  /* Compute the total EEF */

           ee += ee_inner / 2.0;  // remember only half the box

           int imax=200;

         // Loop over annuli from the half-x_width to the diagonal distance
           for (int i=1; i<=imax; i++)
           {

           // Calculate the fraction of this annulus which is
           // contained within the box
             double rout = minor + i * (diagonal - minor)/imax;  
             double rinn = minor + (i-1) * (diagonal - minor)/imax;  

           // Calculate the angle of the circle which lies
           // within the box
             double ang = M_PI / 2 - acos(minor / rout);         
          
           // Subtract a small angle if circle passes outside the box in the
           // major axis, i.e. it is clipped by the box corners
             if (rout>major) { ang -= acos(major / rout); }

             double fraction = 2 * ang / M_PI;  // Half circle

           // Add the extra enclosed energy in this annulus onto the total
           // Could make this more efficient !!
             double eeann = (_psf->encircledEnergy(rout) - 
                                  _psf->encircledEnergy(rinn));
             ee += fraction * eeann;
             //cout<<"ee: "<<i<<" "<<fraction<<" "<<eeann<<" "<<ee<<endl;
           }
         }
     }

    /* Finally correct this value if a subset of the extraction region 
       in burst mode has been used */
     if (_fastmode_factor > 0.0 && _fastmode_factor < 1.0) 
                      ee = ee * _fastmode_factor;

  RETURN ee;
}

/* Method - setBoxShape() - sets the input shape */
void CalPsfModel::setBoxShape(ShapeBasics::ShapeType shape) 
{
  PROC;

     _boxShape = shape;

  ENDPROC;
}

/* Method - setFastModeFactor - sets a factor to scale the ee correction */
void CalPsfModel::setFastModeFactor(double factor)
{
   _fastmode_factor = factor;
}

